@@ -1,10 +1,7 @@
-# This Source Code Form is subject to the terms of the Mozilla Public
-# License, v. 2.0. If a copy of the MPL was not distributed with this
-# file, You can obtain one at http://mozilla.org/MPL/2.0/.
-#
-# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
+# SPDX-FileCopyrightText: 2022 James R. Barlow
+# SPDX-License-Identifier: MPL-2.0
 
-"""Provide classes to stand in for PDF objects
+"""Provide classes to stand in for PDF objects.
 
 The purpose of these is to provide nice-looking classes to allow explicit
 construction of PDF objects and more pythonic idioms and facilitate discovery
@@ -19,10 +16,11 @@ instances of ``pikepdf.Object``, which is a variant container object. The
 class definition is present as an aide for code introspection.
 """
 
-# pylint: disable=unused-import, abstract-method
+from __future__ import annotations
 
+# pylint: disable=unused-import, abstract-method
 from secrets import token_urlsafe
-from typing import TYPE_CHECKING, Any, Iterable, Mapping, Optional, Union, cast
+from typing import TYPE_CHECKING, Any, Iterable, Mapping, cast
 from warnings import warn
 
 from . import _qpdf
@@ -44,7 +42,7 @@ ObjectType.__module__ = __name__
 
 
 class _ObjectMeta(type(Object)):  # type: ignore
-    """Supports instance checking."""
+    """Support instance checking."""
 
     def __instancecheck__(self, instance: Any) -> bool:
         # Note: since this class is a metaclass, self is a class object
@@ -54,7 +52,7 @@ class _ObjectMeta(type(Object)):  # type: ignore
 
 
 class _NameObjectMeta(_ObjectMeta):
-    """Supports usage pikepdf.Name.Whatever -> Name('/Whatever')."""
+    """Support usage pikepdf.Name.Whatever -> Name('/Whatever')."""
 
     def __getattr__(self, attr: str) -> Any:
         if attr.startswith('_') or attr == 'object_type':
@@ -70,7 +68,7 @@ class _NameObjectMeta(_ObjectMeta):
             "modify a Dictionary rather than a Name?"
         )
 
-    def __getitem__(self, item: str) -> 'Name':
+    def __getitem__(self, item: str) -> Name:
         if item.startswith('/'):
             item = item[1:]
         raise TypeError(
@@ -82,7 +80,7 @@ class _NameObjectMeta(_ObjectMeta):
 
 
 class Name(Object, metaclass=_NameObjectMeta):
-    """Constructs a PDF Name object.
+    """Construct a PDF Name object.
 
     Names can be constructed with two notations:
 
@@ -97,7 +95,8 @@ class Name(Object, metaclass=_NameObjectMeta):
 
     object_type = ObjectType.name_
 
-    def __new__(cls, name: Union[str, 'Name']) -> 'Name':
+    def __new__(cls, name: str | Name) -> Name:
+        """Construct a PDF Name."""
         # QPDF_Name::unparse ensures that names are always saved in a UTF-8
         # compatible way, so we only need to guard the input.
         if isinstance(name, bytes):
@@ -107,16 +106,16 @@ class Name(Object, metaclass=_NameObjectMeta):
         return _qpdf._new_name(name)
 
     @classmethod
-    def random(cls, len_: int = 16, prefix: str = '') -> 'Name':
+    def random(cls, len_: int = 16, prefix: str = '') -> Name:
         """Generate a cryptographically strong random, valid PDF Name.
 
         This function uses Python's secrets.token_urlsafe, which returns a
         URL-safe encoded random number of the desired length. An optional
         *prefix* may be prepended. (The encoding is ultimately done with
-        :func:`base64.urlsafe_b64encode`.) Serendipitiously, URL-safe is also
+        :func:`base64.urlsafe_b64encode`.) Serendipitously, URL-safe is also
         PDF-safe.
 
-        When the length paramater is 16 (16 random bytes or 128 bits), the result
+        When the length parameter is 16 (16 random bytes or 128 bits), the result
         is probably globally unique and can be treated as never colliding with
         other names.
 
@@ -127,7 +126,7 @@ class Name(Object, metaclass=_NameObjectMeta):
 
 
 class Operator(Object, metaclass=_ObjectMeta):
-    """Constructs an operator for use in a content stream.
+    """Construct an operator for use in a content stream.
 
     An Operator is one of a limited set of commands that can appear in PDF content
     streams (roughly the mini-language that draws objects, lines and text on a
@@ -141,17 +140,20 @@ class Operator(Object, metaclass=_ObjectMeta):
 
     object_type = ObjectType.operator
 
-    def __new__(cls, name: str) -> 'Operator':
+    def __new__(cls, name: str) -> Operator:
+        """Construct an operator."""
         return cast('Operator', _qpdf._new_operator(name))
 
 
 class String(Object, metaclass=_ObjectMeta):
-    """Constructs a PDF String object."""
+    """Construct a PDF String object."""
 
     object_type = ObjectType.string
 
-    def __new__(cls, s: Union[str, bytes]) -> 'String':
+    def __new__(cls, s: str | bytes) -> String:
         """
+        Construct a PDF String.
+
         Args:
             s: The string to use. String will be encoded for
                 PDF, bytes will be constructed without encoding.
@@ -165,12 +167,14 @@ class String(Object, metaclass=_ObjectMeta):
 
 
 class Array(Object, metaclass=_ObjectMeta):
-    """Constructs a PDF Array object."""
+    """Construct a PDF Array object."""
 
     object_type = ObjectType.array
 
-    def __new__(cls, a: Optional[Union[Iterable, Rectangle]] = None) -> 'Array':
+    def __new__(cls, a: Iterable | Rectangle | None = None) -> Array:
         """
+        Construct a PDF Array.
+
         Args:
             a: An iterable of objects. All objects must be either
                 `pikepdf.Object` or convertible to `pikepdf.Object`.
@@ -178,10 +182,10 @@ class Array(Object, metaclass=_ObjectMeta):
         Return type:
             pikepdf.Array
         """
-
         if isinstance(a, (str, bytes)):
             raise TypeError('Strings cannot be converted to arrays of chars')
-        elif a is None:
+
+        if a is None:
             a = []
         elif isinstance(a, Rectangle):
             return a.as_array()
@@ -191,14 +195,15 @@ class Array(Object, metaclass=_ObjectMeta):
 
 
 class Dictionary(Object, metaclass=_ObjectMeta):
-    """Constructs a PDF Dictionary object."""
+    """Construct a PDF Dictionary object."""
 
     object_type = ObjectType.dictionary
 
-    def __new__(cls, d: Optional[Mapping] = None, **kwargs) -> 'Dictionary':
+    def __new__(cls, d: Mapping | None = None, **kwargs) -> Dictionary:
         """
-        Constructs a PDF Dictionary from either a Python ``dict`` or keyword
-        arguments.
+        Construct a PDF Dictionary.
+
+        Works from either a Python ``dict`` or keyword arguments.
 
         These two examples are equivalent:
 
@@ -232,17 +237,16 @@ class Dictionary(Object, metaclass=_ObjectMeta):
 
 
 class Stream(Object, metaclass=_ObjectMeta):
-    """Constructs a PDF Stream object."""
+    """Construct a PDF Stream object."""
 
     object_type = ObjectType.stream
 
-    def __new__(
-        cls, owner: 'Pdf', data: Optional[bytes] = None, d=None, **kwargs
-    ) -> 'Stream':
+    def __new__(cls, owner: Pdf, data: bytes | None = None, d=None, **kwargs) -> Stream:
         """
-        Create a new stream object, which stores arbitrary binary data and may
-        or may not be compressed. It also may or may not be a page or
-        Form XObject's content stream.
+        Create a new stream object.
+
+        Streams stores arbitrary binary data and may or may not be compressed.
+        It also may or may not be a page or Form XObject's content stream.
 
         A stream dictionary is like a pikepdf.Dictionary or Python dict, except
         it has a binary payload of data attached. The dictionary describes
@@ -283,7 +287,6 @@ class Stream(Object, metaclass=_ObjectMeta):
         .. versionchanged:: 3.0
             Deprecated ``obj`` argument was removed; use ``data``.
         """
-
         if data is None:
             raise TypeError("Must make Stream from binary data")
 
