@@ -1,10 +1,5 @@
-/*
- * This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/.
- *
- * Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
- */
+// SPDX-FileCopyrightText: 2022 James R. Barlow
+// SPDX-License-Identifier: MPL-2.0
 
 #include <cstring>
 #include <cctype>
@@ -15,7 +10,6 @@
 #include <qpdf/QPDFExc.hh>
 #include <qpdf/QPDFObjGen.hh>
 #include <qpdf/QPDFXRefEntry.hh>
-#include <qpdf/PointerHolder.hh>
 #include <qpdf/Buffer.hh>
 #include <qpdf/QPDFObjectHandle.hh>
 #include <qpdf/QPDF.hh>
@@ -47,7 +41,7 @@ PDF semantics dictate that setting a dictionary key to Null deletes the key.
 
     d['/Key'] = None  # would delete /Key
 
-For Python users appears to have an unxpected side effect, so this action is
+For Python users appears to have an unexpected side effect, so this action is
 prohibited. You cannot set keys to None.
 
 pikepdf.String is a "type" that can be converted with str() or bytes() as
@@ -66,21 +60,21 @@ py::size_t list_range_check(QPDFObjectHandle h, int index)
     return static_cast<py::size_t>(index);
 }
 
-inline bool typecode_is_bool(QPDFObject::object_type_e typecode)
+inline bool typecode_is_bool(qpdf_object_type_e typecode)
 {
-    return typecode == QPDFObject::object_type_e::ot_boolean;
+    return typecode == qpdf_object_type_e::ot_boolean;
 }
 
-inline bool typecode_is_int(QPDFObject::object_type_e typecode)
+inline bool typecode_is_int(qpdf_object_type_e typecode)
 {
-    return typecode == QPDFObject::object_type_e::ot_integer;
+    return typecode == qpdf_object_type_e::ot_integer;
 }
 
-inline bool typecode_is_numeric(QPDFObject::object_type_e typecode)
+inline bool typecode_is_numeric(qpdf_object_type_e typecode)
 {
-    return typecode == QPDFObject::object_type_e::ot_integer ||
-           typecode == QPDFObject::object_type_e::ot_real ||
-           typecode == QPDFObject::object_type_e::ot_boolean;
+    return typecode == qpdf_object_type_e::ot_integer ||
+           typecode == qpdf_object_type_e::ot_real ||
+           typecode == qpdf_object_type_e::ot_boolean;
 }
 
 bool objecthandle_equal(QPDFObjectHandle self, QPDFObjectHandle other)
@@ -91,14 +85,11 @@ bool objecthandle_equal(QPDFObjectHandle self, QPDFObjectHandle other)
     if (!self.isInitialized() || !other.isInitialized())
         return false; // LCOV_EXCL_LINE
 
-    // Indirect objects (objid != 0) with the same obj-gen are equal and same owner
-    // are equal (in fact, they are identical; they reference the same underlying
-    // QPDFObject, even if the handles are different).
-    // This lets us compare deeply nested and cyclic structures without recursing
-    // into them.
-    if (self.getObjectID() != 0 && other.getObjectID() != 0 &&
-        self.getOwningQPDF() == other.getOwningQPDF() &&
-        self.getObjGen() == other.getObjGen()) {
+    // If two objects point to the same underlying object, they are equal (in fact,
+    // they are identical; they reference the same underlying QPDFObject, even if the
+    // handles are different). This lets us compare deeply nested and cyclic
+    // structures without recursing into them.
+    if (self.isSameObjectAs(other)) {
         return true;
     }
 
@@ -120,35 +111,35 @@ bool objecthandle_equal(QPDFObjectHandle self, QPDFObjectHandle other)
         return result;
     }
 
-    // Apart from numeric types, disimilar types are never equal
+    // Apart from numeric types, dissimilar types are never equal
     if (self_typecode != other_typecode)
         return false;
 
     switch (self_typecode) {
-    case QPDFObject::object_type_e::ot_null:
+    case qpdf_object_type_e::ot_null:
         return true; // Both must be null
-    case QPDFObject::object_type_e::ot_name:
+    case qpdf_object_type_e::ot_name:
         return self.getName() == other.getName();
-    case QPDFObject::object_type_e::ot_operator:
+    case qpdf_object_type_e::ot_operator:
         return self.getOperatorValue() == other.getOperatorValue();
-    case QPDFObject::object_type_e::ot_string: {
+    case qpdf_object_type_e::ot_string: {
         // We don't know what encoding the string is in
         // This ensures UTF-16 coded ASCII strings will compare equal to
         // UTF-8/ASCII coded.
         return self.getStringValue() == other.getStringValue() ||
                self.getUTF8Value() == other.getUTF8Value();
     }
-    case QPDFObject::object_type_e::ot_array: {
+    case qpdf_object_type_e::ot_array: {
         // Call operator==() on each element of the arrays, meaning this
         // recurses into this function
         return (self.getArrayAsVector() == other.getArrayAsVector());
     }
-    case QPDFObject::object_type_e::ot_dictionary: {
+    case qpdf_object_type_e::ot_dictionary: {
         // Call operator==() on each element of the arrays, meaning this
         // recurses into this function
         return (self.getDictAsMap() == other.getDictAsMap());
     }
-    case QPDFObject::object_type_e::ot_stream: {
+    case qpdf_object_type_e::ot_stream: {
         // Recurse into this function to check if our dictionaries are equal
         if (!objecthandle_equal(self.getDict(), other.getDict()))
             return false;
@@ -158,29 +149,28 @@ bool objecthandle_equal(QPDFObjectHandle self, QPDFObjectHandle other)
         auto self_buffer  = self.getRawStreamData();
         auto other_buffer = other.getRawStreamData();
 
-        // Early out: if underlying QPDF Buffers happen to be the same, the data is the
-        // same
-        if (self_buffer.getPointer() == other_buffer.getPointer())
+        // Early out: if underlying QPDF Buffers happen to be the same, the data is
+        // the same
+        if (self_buffer == other_buffer)
             return true;
         // Early out: if sizes are different, data cannot be the same
-        if (self_buffer.getPointer()->getSize() != other_buffer.getPointer()->getSize())
+        if (self_buffer->getSize() != other_buffer->getSize())
             return false;
 
         // Slow path: memcmp the binary data
-        return 0 == std::memcmp(self_buffer.getPointer()->getBuffer(),
-                        other_buffer.getPointer()->getBuffer(),
-                        self_buffer.getPointer()->getSize());
+        return 0 == std::memcmp(self_buffer->getBuffer(),
+                        other_buffer->getBuffer(),
+                        self_buffer->getSize());
     }
-    case QPDFObject::object_type_e::ot_boolean:
-    case QPDFObject::object_type_e::ot_integer:
-    case QPDFObject::object_type_e::ot_real:
-        // LCOV_EXCL_START
+    // LCOV_EXCL_START
+    case qpdf_object_type_e::ot_boolean:
+    case qpdf_object_type_e::ot_integer:
+    case qpdf_object_type_e::ot_real:
         throw std::logic_error("should have eliminated numeric types by now");
-        // LCOV_EXCL_STOP
     default:
-        break;
+        throw std::logic_error("invalid object type");
     }
-    return false;
+    // LCOV_EXCL_STOP
 }
 
 bool operator==(QPDFObjectHandle self, QPDFObjectHandle other)
@@ -193,7 +183,7 @@ bool operator==(QPDFObjectHandle self, QPDFObjectHandle other)
 bool object_has_key(QPDFObjectHandle h, std::string const &key)
 {
     if (!h.isDictionary() && !h.isStream())
-        throw py::value_error("object is not a dictionary or a stream");
+        throw py::value_error("pikepdf.Object is not a Dictionary or Stream");
     QPDFObjectHandle dict = h.isStream() ? h.getDict() : h;
     return dict.hasKey(key);
 }
@@ -201,7 +191,7 @@ bool object_has_key(QPDFObjectHandle h, std::string const &key)
 bool array_has_item(QPDFObjectHandle haystack, QPDFObjectHandle needle)
 {
     if (!haystack.isArray())
-        throw std::logic_error("object is not an array");
+        throw std::logic_error("pikepdf.Object is not an Array"); // LCOV_EXCL_LINE
 
     auto vec    = haystack.getArrayAsVector();
     auto result = std::find(std::begin(vec), std::end(vec), needle);
@@ -211,7 +201,7 @@ bool array_has_item(QPDFObjectHandle haystack, QPDFObjectHandle needle)
 QPDFObjectHandle object_get_key(QPDFObjectHandle h, std::string const &key)
 {
     if (!h.isDictionary() && !h.isStream())
-        throw py::value_error("object is not a dictionary or a stream");
+        throw py::value_error("pikepdf.Object is not a Dictionary or Stream");
     QPDFObjectHandle dict = h.isStream() ? h.getDict() : h;
     if (!dict.hasKey(key))
         throw py::key_error(key);
@@ -221,7 +211,7 @@ QPDFObjectHandle object_get_key(QPDFObjectHandle h, std::string const &key)
 void object_set_key(QPDFObjectHandle h, std::string const &key, QPDFObjectHandle &value)
 {
     if (!h.isDictionary() && !h.isStream())
-        throw py::value_error("object is not a dictionary or a stream");
+        throw py::value_error("pikepdf.Object is not a Dictionary or Stream");
     if (value.isNull())
         throw py::value_error(
             "PDF Dictionary keys may not be set to None - use 'del' to remove");
@@ -243,7 +233,7 @@ void object_set_key(QPDFObjectHandle h, std::string const &key, QPDFObjectHandle
 void object_del_key(QPDFObjectHandle h, std::string const &key)
 {
     if (!h.isDictionary() && !h.isStream())
-        throw py::value_error("object is not a dictionary or a stream");
+        throw py::value_error("pikepdf.Object is not a Dictionary or Stream");
     if (h.isStream() && key == "/Length") {
         throw py::key_error("/Length may not be deleted");
     }
@@ -263,12 +253,11 @@ std::pair<int, int> object_get_objgen(QPDFObjectHandle h)
     return std::pair<int, int>(objgen.getObj(), objgen.getGen());
 }
 
-PointerHolder<Buffer> get_stream_data(
+std::shared_ptr<Buffer> get_stream_data(
     QPDFObjectHandle &h, qpdf_stream_decode_level_e decode_level)
 {
     try {
-        PointerHolder<Buffer> buf = h.getStreamData(decode_level);
-        return buf;
+        return h.getStreamData(decode_level);
     } catch (const QPDFExc &e) {
         // Make a new exception that has the objgen info, since QPDF's
         // will not
@@ -284,22 +273,22 @@ PointerHolder<Buffer> get_stream_data(
 
 void init_object(py::module_ &m)
 {
-    py::enum_<QPDFObject::object_type_e>(m, "ObjectType")
-        .value("uninitialized", QPDFObject::object_type_e::ot_uninitialized)
-        .value("reserved", QPDFObject::object_type_e::ot_reserved)
-        .value("null", QPDFObject::object_type_e::ot_null)
-        .value("boolean", QPDFObject::object_type_e::ot_boolean)
-        .value("integer", QPDFObject::object_type_e::ot_integer)
-        .value("real", QPDFObject::object_type_e::ot_real)
-        .value("string", QPDFObject::object_type_e::ot_string)
-        .value("name_", QPDFObject::object_type_e::ot_name)
-        .value("array", QPDFObject::object_type_e::ot_array)
-        .value("dictionary", QPDFObject::object_type_e::ot_dictionary)
-        .value("stream", QPDFObject::object_type_e::ot_stream)
-        .value("operator", QPDFObject::object_type_e::ot_operator)
-        .value("inlineimage", QPDFObject::object_type_e::ot_inlineimage);
+    py::enum_<qpdf_object_type_e>(m, "ObjectType")
+        .value("uninitialized", qpdf_object_type_e::ot_uninitialized)
+        .value("reserved", qpdf_object_type_e::ot_reserved)
+        .value("null", qpdf_object_type_e::ot_null)
+        .value("boolean", qpdf_object_type_e::ot_boolean)
+        .value("integer", qpdf_object_type_e::ot_integer)
+        .value("real", qpdf_object_type_e::ot_real)
+        .value("string", qpdf_object_type_e::ot_string)
+        .value("name_", qpdf_object_type_e::ot_name)
+        .value("array", qpdf_object_type_e::ot_array)
+        .value("dictionary", qpdf_object_type_e::ot_dictionary)
+        .value("stream", qpdf_object_type_e::ot_stream)
+        .value("operator", qpdf_object_type_e::ot_operator)
+        .value("inlineimage", qpdf_object_type_e::ot_inlineimage);
 
-    py::class_<Buffer, PointerHolder<Buffer>>(m, "Buffer", py::buffer_protocol())
+    py::class_<Buffer, std::shared_ptr<Buffer>>(m, "Buffer", py::buffer_protocol())
         .def_buffer([](Buffer &b) -> py::buffer_info {
             return py::buffer_info(b.getBuffer(),
                 sizeof(unsigned char),
@@ -381,16 +370,16 @@ void init_object(py::module_ &m)
             [](QPDFObjectHandle &self) -> py::int_ {
                 // Objects which compare equal must have the same hash value
                 switch (self.getTypeCode()) {
-                case QPDFObject::object_type_e::ot_string:
+                case qpdf_object_type_e::ot_string:
                     return py::hash(py::bytes(self.getUTF8Value()));
-                case QPDFObject::object_type_e::ot_name:
+                case qpdf_object_type_e::ot_name:
                     return py::hash(py::bytes(self.getName()));
-                case QPDFObject::object_type_e::ot_operator:
+                case qpdf_object_type_e::ot_operator:
                     return py::hash(py::bytes(self.getOperatorValue()));
-                case QPDFObject::object_type_e::ot_array:
-                case QPDFObject::object_type_e::ot_dictionary:
-                case QPDFObject::object_type_e::ot_stream:
-                case QPDFObject::object_type_e::ot_inlineimage:
+                case qpdf_object_type_e::ot_array:
+                case qpdf_object_type_e::ot_dictionary:
+                case qpdf_object_type_e::ot_stream:
+                case qpdf_object_type_e::ot_inlineimage:
                     throw py::type_error("Can't hash mutable object");
                 default:
                     break;
@@ -400,7 +389,7 @@ void init_object(py::module_ &m)
         .def(
             "__eq__",
             [](QPDFObjectHandle &self, QPDFObjectHandle &other) {
-                return (self == other); // overloaded
+                return objecthandle_equal(self, other);
             },
             py::is_operator())
         .def(
@@ -408,9 +397,9 @@ void init_object(py::module_ &m)
             [](QPDFObjectHandle &self, py::str other) {
                 std::string utf8_other = other.cast<std::string>();
                 switch (self.getTypeCode()) {
-                case QPDFObject::object_type_e::ot_string:
+                case qpdf_object_type_e::ot_string:
                     return self.getUTF8Value() == utf8_other;
-                case QPDFObject::object_type_e::ot_name:
+                case qpdf_object_type_e::ot_name:
                     return self.getName() == utf8_other;
                 default:
                     return false;
@@ -422,9 +411,9 @@ void init_object(py::module_ &m)
             [](QPDFObjectHandle &self, py::bytes other) {
                 std::string bytes_other = other.cast<std::string>();
                 switch (self.getTypeCode()) {
-                case QPDFObject::object_type_e::ot_string:
+                case qpdf_object_type_e::ot_string:
                     return self.getStringValue() == bytes_other;
-                case QPDFObject::object_type_e::ot_name:
+                case qpdf_object_type_e::ot_name:
                     return self.getName() == bytes_other;
                 default:
                     return false;
@@ -444,7 +433,7 @@ void init_object(py::module_ &m)
                     return py::reinterpret_borrow<py::object>(
                         py::handle(Py_NotImplemented));
                 }
-                bool result = (self == q_other);
+                bool result = objecthandle_equal(self, q_other);
                 return py::bool_(result);
             },
             py::is_operator())
@@ -483,15 +472,13 @@ void init_object(py::module_ &m)
             [](QPDFObjectHandle &h, std::string const &key, QPDFObjectHandle &value) {
                 object_set_key(h, key, value);
             },
-            "assign dictionary key to new object",
-            py::keep_alive<1, 3>())
+            "assign dictionary key to new object")
         .def(
             "__setitem__",
             [](QPDFObjectHandle &h, QPDFObjectHandle &name, QPDFObjectHandle &value) {
                 object_set_key(h, name.getName(), value);
             },
-            "assign dictionary key to new object",
-            py::keep_alive<1, 3>())
+            "assign dictionary key to new object")
         .def("__setitem__",
             [](QPDFObjectHandle &h, std::string const &key, py::object pyvalue) {
                 auto value = objecthandle_encode(pyvalue);
@@ -521,12 +508,12 @@ void init_object(py::module_ &m)
                     value = object_get_key(h, key);
                 } catch (const py::key_error &e) {
                     if (std::isupper(name[0]))
-                        throw py::attr_error(e.what());
+                        throw py::attribute_error(e.what());
                     else
-                        throw py::attr_error(name);
+                        throw py::attribute_error(name);
                 } catch (const py::value_error &e) {
                     if (name == std::string("__name__"))
-                        throw py::attr_error(name);
+                        throw py::attribute_error(name);
                     throw;
                 }
                 return value;
@@ -609,9 +596,9 @@ void init_object(py::module_ &m)
             py::return_value_policy::reference_internal)
         .def(
             "keys",
-            [](QPDFObjectHandle h) {
+            [](QPDFObjectHandle &h) {
                 if (h.isStream())
-                    h = h.getDict();
+                    return h.getDict().getKeys();
                 return h.getKeys();
             },
             "For ``pikepdf.Dictionary`` or ``pikepdf.Stream`` objects, obtain the "
@@ -688,7 +675,7 @@ void init_object(py::module_ &m)
                 if (h.isName())
                     return py::bytes(h.getName());
                 if (h.isStream()) {
-                    PointerHolder<Buffer> buf = h.getStreamData();
+                    auto buf = h.getStreamData();
                     // py::bytes will make a copy of the buffer, so releasing is fine
                     return py::bytes((const char *)buf->getBuffer(), buf->getSize());
                 }
@@ -738,22 +725,18 @@ void init_object(py::module_ &m)
             },
             "Extend a pikepdf.Array with an iterable of other objects.")
         .def_property_readonly("is_rectangle",
-            &QPDFObjectHandle::isRectangle,
+            &QPDFObjectHandle::isRectangle, // LCOV_EXCL_LINE
             "Returns True if the object is a rectangle (an array of 4 numbers)")
         .def(
             "get_stream_buffer",
             [](QPDFObjectHandle &h, qpdf_stream_decode_level_e decode_level) {
-                auto phbuf = get_stream_data(h, decode_level);
-                return phbuf;
+                return get_stream_data(h, decode_level);
             },
             "Return a buffer protocol buffer describing the decoded stream.",
             py::arg("decode_level") = qpdf_dl_generalized)
         .def(
             "get_raw_stream_buffer",
-            [](QPDFObjectHandle &h) {
-                PointerHolder<Buffer> phbuf = h.getRawStreamData();
-                return phbuf;
-            },
+            [](QPDFObjectHandle &h) { return h.getRawStreamData(); },
             "Return a buffer protocol buffer describing the raw, encoded stream")
         .def(
             "read_bytes",
@@ -766,7 +749,7 @@ void init_object(py::module_ &m)
         .def(
             "read_raw_bytes",
             [](QPDFObjectHandle &h) {
-                PointerHolder<Buffer> buf = h.getRawStreamData();
+                auto buf = h.getRawStreamData();
                 // py::bytes will make a copy of the buffer, so releasing is fine
                 return py::bytes((const char *)buf->getBuffer(), buf->getSize());
             },
@@ -806,7 +789,6 @@ void init_object(py::module_ &m)
             since it does not take too long for modern CPUs to reconstruct an
             entire PDF. pikepdf will consolidate all incremental updates
             when saving.
-
             )~~~")
         .def_static(
             "parse",
@@ -826,7 +808,7 @@ void init_object(py::module_ &m)
                 return og.getInstructions();
             })
         .def_static("_parse_stream",
-            &QPDFObjectHandle::parseContentStream,
+            &QPDFObjectHandle::parseContentStream, // LCOV_EXCL_LINE
             "Helper for parsing PDF content stream; use "
             "``pikepdf.parse_content_stream``.")
         .def_static("_parse_stream_grouped",
@@ -863,10 +845,13 @@ void init_object(py::module_ &m)
             )~~~")
         .def(
             "to_json",
-            [](QPDFObjectHandle &h, bool dereference = false) -> py::bytes {
-                return h.getJSON(dereference).unparse();
+            [](QPDFObjectHandle &h,
+                bool dereference   = false,
+                int schema_version = 2) -> py::bytes {
+                return h.getJSON(schema_version, dereference).unparse();
             },
-            py::arg("dereference") = false,
+            py::arg("schema_version") = 2,
+            py::arg("dereference")    = false,
             R"~~~(
             Convert to a QPDF JSON representation of the object.
 
@@ -889,12 +874,16 @@ void init_object(py::module_ &m)
             Args:
                 dereference (bool): If True, dereference the object is this is an
                     indirect object.
+                schema_version (int): The version of the JSON schema. Defaults to 2.
 
             Returns:
                 JSON bytestring of object. The object is UTF-8 encoded
                 and may be decoded to a Python str that represents the binary
                 values ``\x00-\xFF`` as ``U+0000`` to ``U+00FF``; that is,
                 it may contain mojibake.
+
+            .. versionchanged:: 6.0
+                Added *schema_version*.
             )~~~"); // end of QPDFObjectHandle bindings
 
     m.def("_new_boolean", &QPDFObjectHandle::newBool, "Construct a PDF Boolean object");
@@ -950,14 +939,11 @@ void init_object(py::module_ &m)
         "can be coerced to PDF objects.");
     m.def(
         "_new_stream",
-        [](std::shared_ptr<QPDF> owner, py::bytes data) {
-            std::string s = data;
-            return QPDFObjectHandle::newStream(owner.get(),
-                data); // This makes a copy of the data
+        [](QPDF &owner, py::bytes data) {
+            // This makes a copy of the data
+            return QPDFObjectHandle::newStream(&owner, data);
         },
-        "Construct a PDF Stream object from binary data",
-        py::keep_alive<0, 1>() // returned object references the owner
-    );
+        "Construct a PDF Stream object from binary data");
     m.def(
         "_new_operator",
         [](const std::string &op) { return QPDFObjectHandle::newOperator(op); },
@@ -978,6 +964,8 @@ void init_object(py::module_ &m)
             Consider instead using :func:`pikepdf.parse_content_stream`.
         )~~~")
         .def(py::init<>(), "You must call ``super.__init__()`` in subclasses.")
+        // LCOV_EXCL_START
+        // coverage misses the virtual function call ::handleObject here.
         .def(
             "handle_object",
             [](QPDFObjectHandle::ParserCallbacks &parsercallbacks,
@@ -990,6 +978,7 @@ void init_object(py::module_ &m)
                 This function will be called back once for each object that is
                 parsed in the content stream.
             )~~~")
+        // LCOV_EXCL_STOP
         .def("handle_eof",
             &QPDFObjectHandle::ParserCallbacks::handleEOF,
             R"~~~(
@@ -1005,12 +994,14 @@ void init_object(py::module_ &m)
 
             Used to expose additional functionality specific to that object type.
         )~~~")
-        .def("__eq__",
+        .def(
+            "__eq__",
             [](QPDFObjectHelper &self, QPDFObjectHelper &other) {
                 // Pages that are copies
                 return objecthandle_equal(
                     self.getObjectHandle(), other.getObjectHandle());
-            })
+            },
+            py::is_operator())
         .def_property_readonly(
             "obj",
             [](QPDFObjectHelper &poh) -> QPDFObjectHandle {
