@@ -1,17 +1,10 @@
-/*
- * This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/.
- *
- * Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
- */
+// SPDX-FileCopyrightText: 2022 James R. Barlow
+// SPDX-License-Identifier: MPL-2.0
 
 #include <qpdf/Constants.h>
 #include <qpdf/Types.h>
 #include <qpdf/DLL.h>
 #include <qpdf/QPDFExc.hh>
-#include <qpdf/PointerHolder.hh>
-#include <qpdf/Buffer.hh>
 #include <qpdf/QPDF.hh>
 #include <qpdf/Pipeline.hh>
 #include <qpdf/QUtil.hh>
@@ -23,7 +16,7 @@
 #include "pipeline.h"
 #include "utils.h"
 
-void Pl_PythonOutput::write(unsigned char *buf, size_t len)
+void Pl_PythonOutput::write(const unsigned char *buf, size_t len)
 {
     py::gil_scoped_acquire gil;
     py::ssize_t so_far = 0;
