@@ -1,10 +1,5 @@
-/*
- * This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/.
- *
- * Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
- */
+// SPDX-FileCopyrightText: 2022 James R. Barlow
+// SPDX-License-Identifier: MPL-2.0
 
 #include <sstream>
 #include <type_traits>
@@ -24,6 +19,7 @@
 #include <qpdf/Pl_Discard.hh>
 #include <qpdf/QPDFAcroFormDocumentHelper.hh>
 #include <qpdf/QPDFEmbeddedFileDocumentHelper.hh>
+#include <qpdf/QPDFLogger.hh>
 
 #include <pybind11/stl.h>
 #include <pybind11/iostream.h>
@@ -50,10 +46,11 @@ void check_stream_is_usable(py::object stream)
     }
 }
 
-void qpdf_basic_settings(QPDF &q)
+void qpdf_basic_settings(QPDF &q) // LCOV_EXCL_LINE
 {
     q.setSuppressWarnings(true);
     q.setImmediateCopyFrom(true);
+    q.setLogger(get_pikepdf_logger());
 }
 
 std::shared_ptr<QPDF> open_pdf(py::object filename_or_stream,
@@ -102,7 +99,8 @@ std::shared_ptr<QPDF> open_pdf(py::object filename_or_stream,
         try {
             auto mmap_input_source =
                 std::make_unique<MmapInputSource>(stream, description, closing_stream);
-            auto input_source = PointerHolder<InputSource>(mmap_input_source.release());
+            auto input_source =
+                std::shared_ptr<InputSource>(mmap_input_source.release());
             py::gil_scoped_release release;
             q->processInputSource(input_source, password.c_str());
             success = true;
@@ -120,7 +118,7 @@ std::shared_ptr<QPDF> open_pdf(py::object filename_or_stream,
     if (!success && access_mode == access_stream) {
         auto stream_input_source = std::make_unique<PythonStreamInputSource>(
             stream, description, closing_stream);
-        auto input_source = PointerHolder<InputSource>(stream_input_source.release());
+        auto input_source = std::shared_ptr<InputSource>(stream_input_source.release());
         py::gil_scoped_release release;
         q->processInputSource(input_source, password.c_str());
         success = true;
@@ -144,6 +142,7 @@ std::shared_ptr<QPDF> open_pdf(py::object filename_or_stream,
         python_warning(
             "A password was provided, but no password was needed to open this PDF.");
     }
+    q->getLogger()->info("Opened a file");
 
     return q;
 }
@@ -172,16 +171,45 @@ void update_xmp_pdfversion(QPDF &q, std::string version)
     impl(pypdf, version);
 }
 
-void setup_encryption(QPDFWriter &w, py::object encryption)
+std::string encryption_password(
+    py::dict encryption, const int encryption_level, const char *keyname)
+{
+    std::string result;
+    if (encryption.contains(keyname)) {
+        if (encryption[keyname].is_none())
+            throw py::value_error(std::string("Encryption ") + keyname +
+                                  " may not be None; use empty string?");
+        if (encryption_level <= 4) {
+            auto success =
+                QUtil::utf8_to_pdf_doc(encryption[keyname].cast<std::string>(), result);
+            if (!success)
+                throw py::value_error("Encryption level is R3/R4 and password is not "
+                                      "encodable as PDFDocEncoding");
+        } else {
+            result = encryption[keyname].cast<std::string>();
+        }
+    }
+    return result;
+}
+
+void setup_encryption(QPDFWriter &w, py::object encryption_obj)
 {
     std::string owner;
     std::string user;
 
+    py::dict encryption;
+
     bool aes      = true;
     bool metadata = true;
     std::map<std::string, bool> allow;
     int encryption_level = 6;
 
+    if (py::isinstance<py::tuple>(encryption_obj)) {
+        encryption = encryption_obj.attr("_asdict")();
+    } else {
+        encryption = encryption_obj;
+    }
+
     if (encryption.contains("R")) {
         if (!py::isinstance<py::int_>(encryption["R"]))
             throw py::type_error("Encryption level 'R' must be an integer");
@@ -194,28 +222,9 @@ void setup_encryption(QPDFWriter &w, py::object encryption)
         python_warning("Encryption R=5 is deprecated");
     }
 
-    if (encryption.contains("owner")) {
-        if (encryption_level <= 4) {
-            auto success =
-                QUtil::utf8_to_pdf_doc(encryption["owner"].cast<std::string>(), owner);
-            if (!success)
-                throw py::value_error("Encryption level is R3/R4 and password is not "
-                                      "encodable as PDFDocEncoding");
-        } else {
-            owner = encryption["owner"].cast<std::string>();
-        }
-    }
-    if (encryption.contains("user")) {
-        if (encryption_level <= 4) {
-            auto success =
-                QUtil::utf8_to_pdf_doc(encryption["user"].cast<std::string>(), user);
-            if (!success)
-                throw py::value_error("Encryption level is R3/R4 and password is not "
-                                      "encodable as PDFDocEncoding");
-        } else {
-            user = encryption["user"].cast<std::string>();
-        }
-    }
+    owner = encryption_password(encryption, encryption_level, "owner");
+    user  = encryption_password(encryption, encryption_level, "user");
+
     if (encryption.contains("allow")) {
         auto pyallow               = encryption["allow"];
         allow["accessibility"]     = pyallow.attr("accessibility").cast<bool>();
@@ -289,7 +298,7 @@ void setup_encryption(QPDFWriter &w, py::object encryption)
             print,
             metadata);
     } else if (encryption_level == 4) {
-        w.setR4EncryptionParameters(user.c_str(),
+        w.setR4EncryptionParametersInsecure(user.c_str(),
             owner.c_str(),
             allow["accessibility"],
             allow["extract"],
@@ -301,7 +310,7 @@ void setup_encryption(QPDFWriter &w, py::object encryption)
             metadata,
             aes);
     } else if (encryption_level == 3) {
-        w.setR3EncryptionParameters(user.c_str(),
+        w.setR3EncryptionParametersInsecure(user.c_str(),
             owner.c_str(),
             allow["accessibility"],
             allow["extract"],
@@ -311,7 +320,7 @@ void setup_encryption(QPDFWriter &w, py::object encryption)
             allow["modify_other"],
             print);
     } else if (encryption_level == 2) {
-        w.setR2EncryptionParameters(user.c_str(),
+        w.setR2EncryptionParametersInsecure(user.c_str(),
             owner.c_str(),
             (print != qpdf_r3p_none),
             allow["modify_assembly"],
@@ -426,23 +435,30 @@ void save_pdf(QPDF &q,
     Pl_PythonOutput output_pipe(description.c_str(), stream);
     w.setOutputPipeline(&output_pipe);
 
-    bool encryption_is_truthy = encryption.equal(py::bool_(true));
+    // Possibilities:
+    // encryption=True -> preserve existing
+    // encryption=False -> remove encryption
+    // encryption=None -> remove encryption
+    // encryption=<dict or NamedTuple> apply new encryption settings
 
-    if (encryption_is_truthy && !q.isEncrypted()) {
-        throw py::value_error(
-            "can't perserve encryption parameters on a file with no encryption");
+    bool removing_encryption =
+        encryption.is_none() || encryption.equal(py::bool_(false));
+    if (!removing_encryption) {
+        if ((normalize_content || !stream_decode_level.is_none())) {
+            throw py::value_error("cannot save with encryption and normalize_content "
+                                  "or stream_decode_level");
+        }
     }
+    bool preserving_encryption = encryption.equal(py::bool_(true));
 
-    if ((encryption_is_truthy || py::isinstance<py::dict>(encryption)) &&
-        (normalize_content || !stream_decode_level.is_none())) {
-        throw py::value_error("cannot save with encryption and normalize_content "
-                              "or stream_decode_level");
-    }
-
-    if (encryption_is_truthy) {
+    if (preserving_encryption) {
+        if (!q.isEncrypted()) {
+            throw py::value_error("can't preserve encryption parameters on "
+                                  "a file with no encryption");
+        }
         w.setPreserveEncryption(true); // Keep existing encryption
-    } else if (encryption.is_none() || encryption.equal(py::bool_(false))) {
-        // Note: encryption.equal(py::bool_(false)) != !encryption_is_truthy
+    } else if (removing_encryption) {
+        // Note: encryption.equal(py::bool_(false)) != !preserving_encryption
         w.setPreserveEncryption(false); // Remove encryption
     } else {
         setup_encryption(w, encryption);
@@ -464,7 +480,7 @@ void save_pdf(QPDF &q,
     }
 
     if (!progress.is_none()) {
-        auto reporter = PointerHolder<QPDFWriter::ProgressReporter>(
+        auto reporter = std::shared_ptr<QPDFWriter::ProgressReporter>(
             new PikeProgressReporter(progress));
         w.registerProgressReporter(reporter);
     }
@@ -510,7 +526,7 @@ void init_qpdf(py::module_ &m)
                 qpdf_basic_settings(*q);
                 return q;
             },
-            "Create a new empty PDF from stratch.")
+            "Create a new empty PDF from scratch.")
         .def_static("_open",
             open_pdf,
             py::arg("filename_or_stream"),
@@ -536,7 +552,7 @@ void init_qpdf(py::module_ &m)
         .def_property_readonly("extension_level", &QPDF::getExtensionLevel)
         .def_property_readonly("Root", &QPDF::getRoot, "The /Root object of the PDF.")
         .def_property_readonly("trailer",
-            &QPDF::getTrailer,
+            &QPDF::getTrailer, // LCOV_EXCL_LINE
             R"~~~(
             Provides access to the PDF trailer object.
 
@@ -627,9 +643,7 @@ void init_qpdf(py::module_ &m)
                 first (bool): If True, prepend this before the first page; if False append after last page
             )~~~",
             py::arg("page"),
-            py::arg("first") = false,
-            py::keep_alive<1, 2>())
-        .def("_add_page_at", &QPDF::addPageAt, py::keep_alive<1, 2>())
+            py::arg("first") = false)
         .def("_remove_page", &QPDF::removePage)
         .def(
             "remove_unreferenced_resources",
@@ -686,7 +700,7 @@ void init_qpdf(py::module_ &m)
             Return type:
                 pikepdf.Object
             )~~~",
-            py::return_value_policy::reference_internal,
+            py::return_value_policy::reference_internal, // LCOV_EXCL_LINE
             py::arg("objgen"))
         .def(
             "get_object",
@@ -709,7 +723,7 @@ void init_qpdf(py::module_ &m)
             After deleting content from a PDF such as pages, objects related
             to that page, such as images on the page, may still be present.
 
-            Retun type:
+            Return type:
                 pikepdf._qpdf._ObjectList
             )~~~",
             py::return_value_policy::reference_internal)
@@ -756,6 +770,10 @@ void init_qpdf(py::module_ &m)
             R"~~~(
             Copy an ``Object`` from a foreign ``Pdf`` to this one.
 
+            If you want to copy a page from one PDF to another, use:
+            ``pdf_b.pages[0] = pdf_a.pages[0]``. That interface accounts for the
+            complexity of copying pages.
+
             This function is used to copy a :class:`pikepdf.Object` that is owned by
             some other ``Pdf`` into this one. This is performs a deep (recursive) copy
             and preserves circular references that may exist in the foreign object.
@@ -779,13 +797,11 @@ void init_qpdf(py::module_ &m)
                 Error messages improved.
             )~~~",
             py::return_value_policy::reference_internal,
-            py::keep_alive<1, 2>(),
             py::arg("h"))
         .def("copy_foreign",
             [](QPDF &q, QPDFPageObjectHelper &poh) -> QPDFPageObjectHelper {
-                deprecation_warning("copy_foreign() called on pikepdf.Page - use "
-                                    "Pdf.pages interface instead");
-                return QPDFPageObjectHelper(q.copyForeignObject(poh.getObjectHandle()));
+                throw py::notimpl_error("Use pikepdf.Pdf.pages interface to copy "
+                                        "pages from one PDF to another.");
             })
         .def("_replace_object",
             [](QPDF &q, std::pair<int, int> objgen, QPDFObjectHandle &h) {
@@ -811,8 +827,13 @@ void init_qpdf(py::module_ &m)
                     w.write();
                 } catch (py::error_already_set &e) {
                     auto cls_dependency_error =
-                        py::module_::import("pikepdf.models.image").attr("DependencyError");
+                        py::module_::import("pikepdf._exceptions")
+                            .attr("DependencyError");
                     if (e.matches(cls_dependency_error)) {
+                        python_warning(
+                            "pikepdf is missing some specialized decoders "
+                            "(probably JBIG2) so not all stream contents can be "
+                            "tested.");
                         w.setDecodeLevel(qpdf_dl_generalized);
                         w.write();
                     } else {
