@@ -1,10 +1,5 @@
-/*
- * This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/.
- *
- * Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
- */
+// SPDX-FileCopyrightText: 2022 James R. Barlow
+// SPDX-License-Identifier: MPL-2.0
 
 #include "pikepdf.h"
 #include "qpdf_pagelist.h"
@@ -262,7 +257,6 @@ void init_pagelist(py::module_ &m)
                 auto uindex = uindex_from_index(pl, index);
                 pl.insert_page(uindex, obj);
             },
-            py::keep_alive<1, 3>(),
             R"~~~(
             Insert a page at the specified location.
 
@@ -270,7 +264,7 @@ void init_pagelist(py::module_ &m)
                 index (int): location at which to insert page, 0-based indexing
                 obj (pikepdf.Object): page object to insert
             )~~~",
-            py::arg("index"),
+            py::arg("index"), // LCOV_EXCL_LINE
             py::arg("obj"))
         .def(
             "reverse",
@@ -286,7 +280,6 @@ void init_pagelist(py::module_ &m)
         .def(
             "append",
             [](PageList &pl, py::object page) { pl.insert_page(pl.count(), page); },
-            py::keep_alive<1, 2>(),
             "Add another page to the end.",
             py::arg("page"))
         .def(
@@ -300,7 +293,6 @@ void init_pagelist(py::module_ &m)
                     pl.insert_page(pl.count(), other.get_page_obj(i));
                 }
             },
-            py::keep_alive<1, 2>(),
             "Extend the ``Pdf`` by adding pages from another ``Pdf.pages``.",
             py::arg("other"))
         .def(
@@ -314,7 +306,6 @@ void init_pagelist(py::module_ &m)
                     ++it;
                 }
             },
-            py::keep_alive<1, 2>(),
             "Extend the ``Pdf`` by adding pages from an iterable of pages.",
             py::arg("iterable"))
         .def(
