@@ -1,10 +1,5 @@
-/*
- * This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/.
- *
- * Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
- */
+// SPDX-FileCopyrightText: 2022 James R. Barlow
+// SPDX-License-Identifier: MPL-2.0
 
 #pragma once
 
@@ -16,7 +11,7 @@
 
 void init_pagelist(py::module_ &m);
 
-class PageList {
+class PageList { // LCOV_EXCL_LINE
 public:
     PageList(std::shared_ptr<QPDF> q, py::size_t iterpos = 0)
         : iterpos(iterpos), qpdf(q){};
