@@ -1,11 +1,16 @@
+# SPDX-FileCopyrightText: 2022 James R. Barlow
+# SPDX-License-Identifier: CC0-1.0
+
+from __future__ import annotations
+
 from decimal import Decimal, getcontext
 
 import pytest
 
 import pikepdf
-from pikepdf import _qpdf as qpdf
+from pikepdf.settings import get_decimal_precision, set_decimal_precision
 
-encode = qpdf._encode
+encode = pikepdf._qpdf._encode
 
 # pylint: disable=redefined-outer-name
 
@@ -17,26 +22,26 @@ def test_decimal_precision():
 
 def test_decimal_change_precision():
     d = Decimal('0.1234567890123456789')
-    saved = qpdf.get_decimal_precision()
+    saved = get_decimal_precision()
     try:
-        qpdf.set_decimal_precision(10)
+        set_decimal_precision(10)
         assert str(encode(d)) == '0.1234567890'
-        assert qpdf.get_decimal_precision() == 10
+        assert get_decimal_precision() == 10
     finally:
-        qpdf.set_decimal_precision(saved)
+        set_decimal_precision(saved)
 
 
 def test_decimal_independent_of_app():
     d = Decimal('0.1234567890123456789')
-    pikepdf_prec = qpdf.get_decimal_precision()
+    pikepdf_prec = get_decimal_precision()
     decimal_prec = getcontext().prec
     try:
         getcontext().prec = 6
-        qpdf.set_decimal_precision(8)
+        set_decimal_precision(8)
         assert str(encode(d)) == '0.12345679'
-        assert qpdf.get_decimal_precision() != 6
+        assert get_decimal_precision() != 6
     finally:
-        qpdf.set_decimal_precision(pikepdf_prec)
+        set_decimal_precision(pikepdf_prec)
         getcontext().prec = decimal_prec
 
 
