@@ -1,22 +1,28 @@
+# SPDX-FileCopyrightText: 2022 James R. Barlow
+# SPDX-License-Identifier: CC0-1.0
+
 """
 Testing focused on pikepdf.Pdf
 """
 
+from __future__ import annotations
+
 import locale
-import os
 import shutil
-import sys
 import zlib
 from contextlib import nullcontext
 from io import BytesIO, StringIO
 from os import fspath
 from pathlib import Path
+from subprocess import CalledProcessError
 from unittest.mock import Mock
 
 import pytest
+from conftest import fails_if_pypy
 
 import pikepdf
 from pikepdf import Name, PasswordError, Pdf, PdfError, Stream
+from pikepdf._exceptions import DependencyError
 
 # pylint: disable=redefined-outer-name
 
@@ -24,7 +30,7 @@ from pikepdf import Name, PasswordError, Pdf, PdfError, Stream
 @pytest.fixture
 def trivial(resources):
     with Pdf.open(
-        resources / 'pal-1bit-trivial.pdf', access_mode=pikepdf.AccessMode.mmap
+        resources / 'pal-1bit-trivial.pdf'  # , access_mode=pikepdf.AccessMode.mmap
     ) as pdf:
         yield pdf
 
@@ -335,6 +341,23 @@ def test_check(resources):
         assert 'parse error while reading' in problems[0]
 
 
+@fails_if_pypy
+def test_check_specialized_decoder_fallback(resources, monkeypatch):
+    class MissingJBIG2Decoder(pikepdf.jbig2.JBIG2DecoderInterface):
+        def check_available(self):
+            raise DependencyError('jbig2dec')
+
+        def decode_jbig2(self, jbig2: bytes, jbig2_globals: bytes) -> bytes:
+            raise CalledProcessError(1, 'jbig2dec')
+
+    monkeypatch.setattr(pikepdf.jbig2, 'get_decoder', MissingJBIG2Decoder)
+
+    with pikepdf.open(resources / 'jbig2.pdf') as pdf:
+        with pytest.warns(UserWarning, match=r".*missing some specialized.*"):
+            problems = pdf.check()
+        assert len(problems) == 0
+
+
 def test_repr(trivial):
     assert repr(trivial).startswith('<')
 
@@ -361,17 +384,17 @@ def test_invalid_flate_compression_level():
     # and will change subsequent test results, so just ping it with an invalid
     # value to get partial code coverage.
     with pytest.raises(ValueError):
-        pikepdf._qpdf.set_flate_compression_level(99)
+        pikepdf.settings.set_flate_compression_level(99)
 
 
 def test_flate_compression_level():
     # While this function affects global state, we can test it safely because
     # setting the value to -1 restores the default.
     try:
-        pikepdf._qpdf.set_flate_compression_level(0)
-        pikepdf._qpdf.set_flate_compression_level(9)
+        pikepdf.settings.set_flate_compression_level(0)
+        pikepdf.settings.set_flate_compression_level(9)
     finally:
-        pikepdf._qpdf.set_flate_compression_level(-1)
+        pikepdf.settings.set_flate_compression_level(-1)
 
 
 def test_set_access_default_mmap():
