@@ -1,7 +1,12 @@
+# SPDX-FileCopyrightText: 2022 James R. Barlow
+# SPDX-License-Identifier: CC0-1.0
+
 """
 A bunch of quick tests that confirm nothing is horribly wrong
 """
 
+from __future__ import annotations
+
 import ast
 import gc
 from contextlib import suppress
@@ -9,7 +14,12 @@ from pathlib import Path
 from shutil import copy
 
 import pytest
-import tomli
+
+try:
+    import tomllib
+except ModuleNotFoundError:
+    import tomli as tomllib  # type: ignore
+
 from packaging.version import Version
 
 import pikepdf
@@ -20,7 +30,7 @@ def test_minimum_qpdf_version():
     from pikepdf import _qpdf
 
     with open(Path(__file__).parent / '../pyproject.toml', 'rb') as f:
-        pyproject_toml = tomli.load(f)
+        pyproject_toml = tomllib.load(f)
     toml_env = pyproject_toml['tool']['cibuildwheel']['environment']
 
     assert Version(_qpdf.qpdf_version()) >= Version(toml_env['QPDF_MIN_VERSION'])
